@@ -19,7 +19,7 @@ Other enhancements, Florent Guillaume <fg@nuxeo.com>
 Reworked, Jeroen Ruigrok van der Werven <asmodai@in-nomine.org>
 Jeff Moreland <hou5e@hotmail.com>
 
-$Id: bugzilla2trac.py 9298 2010-02-27 09:59:34Z cboos $
+$Id: bugzilla2trac.py 10009 2010-08-23 22:06:29Z rblank $
 """
 
 import re
@@ -939,7 +939,7 @@ def log(msg):
 
 def datetime2epoch(dt) :
     import time
-    return time.mktime(dt.timetuple())
+    return time.mktime(dt.timetuple()) * 1000000
 
 def usage():
     print """bugzilla2trac - Imports a bug database from Bugzilla into Trac.
