@@ -1,7 +1,7 @@
 """Inserts the current time (in seconds) into the wiki page."""
 
 revision = "$Rev: 6326 $"
-url = "$URL: https://svn.edgewall.org/repos/trac/tags/trac-0.12/sample-plugins/Timestamp.py $"
+url = "$URL: https://svn.edgewall.org/repos/trac/tags/trac-0.12.1/sample-plugins/Timestamp.py $"
 
 #
 # The following shows the code for macro, old-style.
