@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2003-2008 Edgewall Software
+# Copyright (C) 2003-2009 Edgewall Software
 # Copyright (C) 2003-2005 Jonas Borgström <jonas@edgewall.com>
 # Copyright (C) 2005 Christopher Lenz <cmlenz@gmx.de>
 # All rights reserved.
@@ -17,26 +17,29 @@
 #         Christopher Lenz <cmlenz@gmx.de>
 
 from datetime import datetime
-import os
+import os.path
 import re
 import shutil
-import time
+import sys
 import unicodedata
 
 from genshi.builder import tag
 
-from trac import perm, util
+from trac.admin import AdminCommandError, IAdminCommandProvider, PrefixList, \
+                       console_datetime_format, get_dir_list
 from trac.config import BoolOption, IntOption
 from trac.core import *
 from trac.env import IEnvironmentSetupParticipant
 from trac.mimeview import *
-from trac.perm import PermissionError, PermissionSystem, IPermissionPolicy
+from trac.perm import PermissionError, IPermissionPolicy
 from trac.resource import *
 from trac.search import search_to_sql, shorten_result
-from trac.util import get_reporter_id, create_unique_file, content_disposition
-from trac.util.datefmt import to_timestamp, utc
-from trac.util.text import unicode_quote, unicode_unquote, pretty_size
-from trac.util.translation import _
+from trac.util import get_reporter_id, create_unique_file
+from trac.util.datefmt import format_datetime, from_utimestamp, \
+                              to_datetime, to_utimestamp, utc
+from trac.util.text import exception_to_unicode, pretty_size, print_table, \
+                           unicode_quote, unicode_unquote
+from trac.util.translation import _, tag_
 from trac.web import HTTPBadRequest, IRequestHandler
 from trac.web.chrome import add_link, add_stylesheet, add_ctxtnav, \
                             INavigationContributor
@@ -59,6 +62,9 @@ class IAttachmentChangeListener(Interface):
     def attachment_deleted(attachment):
         """Called when an attachment is deleted."""
 
+    def attachment_reparented(attachment, old_parent_realm, old_parent_id):
+        """Called when an attachment is reparented."""
+
 
 class IAttachmentManipulator(Interface):
     """Extension point interface for components that need to manipulate
@@ -134,78 +140,113 @@ class Attachment(object):
         if not db:
             db = self.env.get_db_cnx()
         cursor = db.cursor()
-        cursor.execute("SELECT filename,description,size,time,author,ipnr "
-                       "FROM attachment WHERE type=%s AND id=%s "
-                       "AND filename=%s ORDER BY time",
-                       (self.parent_realm, unicode(self.parent_id), filename))
+        cursor.execute("""
+            SELECT filename,description,size,time,author,ipnr FROM attachment
+            WHERE type=%s AND id=%s AND filename=%s
+            ORDER BY time
+            """, (self.parent_realm, unicode(self.parent_id), filename))
         row = cursor.fetchone()
         cursor.close()
         if not row:
             self.filename = filename
             raise ResourceNotFound(_("Attachment '%(title)s' does not exist.",
-                                     title=self.title), _('Invalid Attachment'))
+                                     title=self.title),
+                                   _('Invalid Attachment'))
         self.filename = row[0]
         self.description = row[1]
         self.size = row[2] and int(row[2]) or 0
-        time = row[3] and int(row[3]) or 0
-        self.date = datetime.fromtimestamp(time, utc)
+        self.date = from_utimestamp(row[3])
         self.author = row[4]
         self.ipnr = row[5]
 
-    def _get_path(self):
-        path = os.path.join(self.env.path, 'attachments', self.parent_realm,
-                            unicode_quote(self.parent_id))
-        if self.filename:
-            path = os.path.join(path, unicode_quote(self.filename))
+    def _get_path(self, parent_realm, parent_id, filename):
+        path = os.path.join(self.env.path, 'attachments', parent_realm,
+                            unicode_quote(parent_id))
+        if filename:
+            path = os.path.join(path, unicode_quote(filename))
         return os.path.normpath(path)
-    path = property(_get_path)
+    
+    @property
+    def path(self):
+        return self._get_path(self.parent_realm, self.parent_id, self.filename)
 
-    def _get_title(self):
-        return '%s:%s: %s' % (self.parent_realm, 
-                              self.parent_id, self.filename)
-    title = property(_get_title)
+    @property
+    def title(self):
+        return '%s:%s: %s' % (self.parent_realm, self.parent_id, self.filename)
 
     def delete(self, db=None):
         assert self.filename, 'Cannot delete non-existent attachment'
-        if not db:
-            db = self.env.get_db_cnx()
-            handle_ta = True
-        else:
-            handle_ta = False
 
-        cursor = db.cursor()
-        cursor.execute("DELETE FROM attachment WHERE type=%s AND id=%s "
-                       "AND filename=%s", (self.parent_realm, self.parent_id,
-                       self.filename))
-        if os.path.isfile(self.path):
-            try:
-                os.unlink(self.path)
-            except OSError:
-                self.env.log.error('Failed to delete attachment file %s',
-                                   self.path, exc_info=True)
-                if handle_ta:
-                    db.rollback()
-                raise TracError(_('Could not delete attachment'))
+        @self.env.with_transaction(db)
+        def do_delete(db):
+            cursor = db.cursor()
+            cursor.execute("DELETE FROM attachment WHERE type=%s AND id=%s "
+                           "AND filename=%s",
+                           (self.parent_realm, self.parent_id, self.filename))
+            if os.path.isfile(self.path):
+                try:
+                    os.unlink(self.path)
+                except OSError, e:
+                    self.env.log.error('Failed to delete attachment '
+                                       'file %s: %s',
+                                       self.path,
+                                       exception_to_unicode(e, traceback=True))
+                    raise TracError(_('Could not delete attachment'))
 
         self.env.log.info('Attachment removed: %s' % self.title)
-        if handle_ta:
-            db.commit()
 
         for listener in AttachmentModule(self.env).change_listeners:
             listener.attachment_deleted(self)
 
+    def reparent(self, new_realm, new_id):
+        assert self.filename, 'Cannot reparent non-existent attachment'
+        new_id = unicode(new_id)
+        
+        @self.env.with_transaction()
+        def do_reparent(db):
+            cursor = db.cursor()
+            new_path = self._get_path(new_realm, new_id, self.filename)
+            if os.path.exists(new_path):
+                raise TracError(_('Cannot reparent attachment "%(att)s" as '
+                                  'it already exists in %(realm)s:%(id)s', 
+                                  att=self.filename, realm=new_realm,
+                                  id=new_id))
+            cursor.execute("""
+                UPDATE attachment SET type=%s, id=%s
+                WHERE type=%s AND id=%s AND filename=%s
+                """, (new_realm, new_id, self.parent_realm, self.parent_id,
+                      self.filename))
+            dirname = os.path.dirname(new_path)
+            if not os.path.exists(dirname):
+                os.makedirs(dirname)
+            if os.path.isfile(self.path):
+                try:
+                    os.rename(self.path, new_path)
+                except OSError, e:
+                    self.env.log.error('Failed to move attachment file %s: %s',
+                                       self.path,
+                                       exception_to_unicode(e, traceback=True))
+                    raise TracError(_('Could not reparent attachment %(name)s',
+                                      name=self.filename))
+
+        old_realm, old_id = self.parent_realm, self.parent_id
+        self.parent_realm, self.parent_id = new_realm, new_id
+        self.resource = Resource(new_realm, new_id).child('attachment',
+                                                          self.filename)
+        
+        self.env.log.info('Attachment reparented: %s' % self.title)
+
+        for listener in AttachmentModule(self.env).change_listeners:
+            if hasattr(listener, 'attachment_reparented'):
+                listener.attachment_reparented(self, old_realm, old_id)
 
     def insert(self, filename, fileobj, size, t=None, db=None):
-        # FIXME: `t` should probably be switched to `datetime` too
-        if not db:
-            db = self.env.get_db_cnx()
-            handle_ta = True
-        else:
-            handle_ta = False
-
         self.size = size and int(size) or 0
-        timestamp = int(t or time.time())
-        self.date = datetime.fromtimestamp(timestamp, utc)
+        if t is None:
+            t = datetime.now(utc)
+        elif not isinstance(t, datetime): # Compatibility with 0.11
+            t = to_datetime(t, utc)
+        self.date = t
 
         # Make sure the path to the attachment is inside the environment
         # attachments directory
@@ -225,27 +266,27 @@ class Attachment(object):
             basename = os.path.basename(path).encode('ascii')
             filename = unicode_unquote(basename)
 
-            cursor = db.cursor()
-            cursor.execute("INSERT INTO attachment "
-                           "VALUES (%s,%s,%s,%s,%s,%s,%s,%s)",
-                           (self.parent_realm, self.parent_id, filename,
-                            self.size, timestamp, self.description,
-                            self.author, self.ipnr))
-            shutil.copyfileobj(fileobj, targetfile)
-            self.resource.id = self.filename = filename
-
-            self.env.log.info('New attachment: %s by %s', self.title,
-                              self.author)
-
-            if handle_ta:
-                db.commit()
-
-            for listener in AttachmentModule(self.env).change_listeners:
-                listener.attachment_added(self)
+            @self.env.with_transaction(db)
+            def do_insert(db):
+                cursor = db.cursor()
+                cursor.execute("INSERT INTO attachment "
+                               "VALUES (%s,%s,%s,%s,%s,%s,%s,%s)",
+                               (self.parent_realm, self.parent_id, filename,
+                                self.size, to_utimestamp(t), self.description,
+                                self.author, self.ipnr))
+                shutil.copyfileobj(fileobj, targetfile)
+                self.resource.id = self.filename = filename
 
+                self.env.log.info('New attachment: %s by %s', self.title,
+                                  self.author)
         finally:
             targetfile.close()
 
+        for listener in AttachmentModule(self.env).change_listeners:
+            listener.attachment_added(self)
+
+
+    @classmethod
     def select(cls, env, parent_realm, parent_id, db=None):
         if not db:
             db = env.get_db_cnx()
@@ -253,37 +294,50 @@ class Attachment(object):
         cursor.execute("SELECT filename,description,size,time,author,ipnr "
                        "FROM attachment WHERE type=%s AND id=%s ORDER BY time",
                        (parent_realm, unicode(parent_id)))
-        for filename,description,size,time,author,ipnr in cursor:
+        for filename, description, size, time, author, ipnr in cursor:
             attachment = Attachment(env, parent_realm, parent_id)
             attachment.filename = filename
             attachment.description = description
             attachment.size = size and int(size) or 0
-            time = time and int(time) or 0
-            attachment.date = datetime.fromtimestamp(time, utc)
+            attachment.date = from_utimestamp(time or 0)
             attachment.author = author
             attachment.ipnr = ipnr
             yield attachment
 
-    def delete_all(cls, env, parent_realm, parent_id, db):
-        """Delete all attachments of a given resource.
-
-        As this is usually done while deleting the parent resource,
-        the `db` argument is ''not'' optional here.
-        """
-        attachment_dir = None
-        for attachment in list(cls.select(env, parent_realm, parent_id, db)):
-            attachment_dir = os.path.dirname(attachment.path)
-            attachment.delete(db)
-        if attachment_dir:
+    @classmethod
+    def delete_all(cls, env, parent_realm, parent_id, db=None):
+        """Delete all attachments of a given resource."""
+        attachment_dir = [None]
+        @env.with_transaction(db)
+        def do_delete(db):
+            for attachment in list(cls.select(env, parent_realm, parent_id,
+                                              db)):
+                attachment_dir[0] = os.path.dirname(attachment.path)
+                attachment.delete()
+        if attachment_dir[0]:
             try:
-                os.rmdir(attachment_dir)
-            except OSError:
-                env.log.error("Can't delete attachment directory %s",
-                              attachment_dir, exc_info=True)
-            
-    select = classmethod(select)
-    delete_all = classmethod(delete_all)
+                os.rmdir(attachment_dir[0])
+            except OSError, e:
+                env.log.error("Can't delete attachment directory %s: %s",
+                    attachment_dir[0], exception_to_unicode(e, traceback=True))
 
+    @classmethod
+    def reparent_all(cls, env, parent_realm, parent_id, new_realm, new_id):
+        """Reparent all attachments of a given resource to another resource."""
+        attachment_dir = [None]
+        @env.with_transaction()
+        def do_reparent(db):
+            for attachment in list(cls.select(env, parent_realm, parent_id,
+                                              db)):
+                attachment_dir = os.path.dirname(attachment.path)
+                attachment.reparent(new_realm, new_id)
+        if attachment_dir[0]:
+            try:
+                os.rmdir(attachment_dir[0])
+            except OSError, e:
+                env.log.error("Can't delete attachment directory %s: %s",
+                    attachment_dir[0], exception_to_unicode(e, traceback=True))
+            
     def open(self):
         self.env.log.debug('Trying to open attachment at %s', self.path)
         try:
@@ -447,9 +501,9 @@ class AttachmentModule(Component):
                        "  FROM attachment "
                        "  WHERE time > %s AND time < %s "
                        "        AND type = %s",
-                       (to_timestamp(start), to_timestamp(stop), realm))
+                       (to_utimestamp(start), to_utimestamp(stop), realm))
         for realm, id, filename, ts, description, author in cursor:
-            time = datetime.fromtimestamp(ts, utc)
+            time = from_utimestamp(ts)
             yield ('created', realm, id, filename, time, description, author)
 
     def get_timeline_events(self, req, resource_realm, start, stop):
@@ -471,8 +525,9 @@ class AttachmentModule(Component):
         elif field == 'title':
             name = get_resource_name(self.env, attachment.parent)
             title = get_resource_summary(self.env, attachment.parent)
-            return tag(tag.em(os.path.basename(attachment.id)),
-                       _(" attached to "), tag.em(name, title=title))
+            return tag_("%(attachment)s attached to %(resource)s",
+                        attachment=tag.em(os.path.basename(attachment.id)),
+                        resource=tag.em(name, title=title))
         elif field == 'description':
             return format_to(self.env, None, context(attachment.parent), descr)
    
@@ -497,7 +552,7 @@ class AttachmentModule(Component):
             if 'ATTACHMENT_VIEW' in req.perm(attachment):
                 yield (get_resource_url(self.env, attachment, req.href),
                        get_resource_shortname(self.env, attachment),
-                       datetime.fromtimestamp(time, utc), author,
+                       from_utimestamp(time), author,
                        shorten_result(desc, terms))
     
     # IResourceManager methods
@@ -511,6 +566,8 @@ class AttachmentModule(Component):
         A `format` keyword argument equal to `'raw'` will be converted
         to the raw-attachment prefix.
         """
+        if not resource.parent:
+            return None
         format = kwargs.get('format')
         prefix = 'attachment'
         if format == 'raw':
@@ -526,6 +583,8 @@ class AttachmentModule(Component):
             return href(prefix, parent_href, resource.id, **kwargs)
 
     def get_resource_description(self, resource, format=None, **kwargs):
+        if not resource.parent:
+            return _("Unparented attachment %(id)s", id=resource.id)
         if format == 'compact':
             return '%s (%s)' % (resource.id,
                     get_resource_name(self.env, resource.parent))
@@ -538,14 +597,25 @@ class AttachmentModule(Component):
             return _("Attachments of %(parent)s",
                      parent=get_resource_name(self.env, resource.parent))
 
+    def resource_exists(self, resource):
+        try:
+            attachment = Attachment(self.env, resource)
+            return os.path.exists(attachment.path)
+        except ResourceNotFound:
+            return False
+
     # Internal methods
 
     def _do_save(self, req, attachment):
         req.perm(attachment.resource).require('ATTACHMENT_CREATE')
+        parent_resource = attachment.resource.parent
+        if not resource_exists(self.env, parent_resource):
+            raise ResourceNotFound(
+                _("%(parent)s doesn't exist, can't create attachment",
+                  parent=get_resource_name(self.env, parent_resource)))
 
         if 'cancel' in req.args:
-            req.redirect(get_resource_url(self.env, attachment.resource.parent,
-                                          req.href))
+            req.redirect(get_resource_url(self.env, parent_resource, req.href))
 
         upload = req.args['attachment']
         if not hasattr(upload, 'filename') or not upload.filename:
@@ -584,20 +654,29 @@ class AttachmentModule(Component):
             for field, message in manipulator.validate_attachment(req,
                                                                   attachment):
                 if field:
-                    raise InvalidAttachment(_('Attachment field %(field)s is '
-                                              'invalid: %(message)s',
-                                              field=field, message=message))
+                    raise InvalidAttachment(
+                        _('Attachment field %(field)s is invalid: %(message)s',
+                          field=field, message=message))
                 else:
-                    raise InvalidAttachment(_('Invalid attachment: %(message)s',
-                                              message=message))
+                    raise InvalidAttachment(
+                        _('Invalid attachment: %(message)s', message=message))
 
         if req.args.get('replace'):
             try:
                 old_attachment = Attachment(self.env,
                                             attachment.resource(id=filename))
-                if not (old_attachment.author and req.authname \
-                        and old_attachment.author == req.authname):
-                    req.perm(attachment.resource).require('ATTACHMENT_DELETE')
+                if not (req.authname and req.authname != 'anonymous' \
+                        and old_attachment.author == req.authname) \
+                   and 'ATTACHMENT_DELETE' \
+                                        not in req.perm(attachment.resource):
+                    raise PermissionError(msg=_("You don't have permission to "
+                        "replace the attachment %(name)s. You can only "
+                        "replace your own attachments. Replacing other's "
+                        "attachments requires ATTACHMENT_DELETE permission.",
+                        name=filename))
+                if (not attachment.description.strip() and
+                    old_attachment.description):
+                    attachment.description = old_attachment.description
                 old_attachment.delete()
             except TracError:
                 pass # don't worry if there's nothing to replace
@@ -632,7 +711,6 @@ class AttachmentModule(Component):
             'attachment': attachment, 'max_size': self.max_size}
 
     def _render_list(self, req, parent):
-        attachment = parent.child('attachment')
         data = {
             'mode': 'list',
             'attachment': None, # no specific attachment
@@ -670,7 +748,7 @@ class AttachmentModule(Component):
                     # XSS attacks
                     req.send_header('Content-Disposition', 'attachment')
                 if format == 'txt':
-                      mime_type = 'text/plain'
+                    mime_type = 'text/plain'
                 elif not mime_type:
                     mime_type = 'application/octet-stream'
                 if 'charset=' not in mime_type:
@@ -723,14 +801,24 @@ class AttachmentModule(Component):
         if attachment and 'ATTACHMENT_VIEW' in formatter.perm(attachment):
             try:
                 model = Attachment(self.env, attachment)
-                format = None
+                raw_href = get_resource_url(self.env, attachment,
+                                            formatter.href, format='raw')
                 if ns.startswith('raw'):
-                    format = 'raw'
-                href = get_resource_url(self.env, attachment, formatter.href,
-                                        format=format)
-                return tag.a(label, class_='attachment', href=href + params,
-                             title=get_resource_name(self.env, attachment))
-            except ResourceNotFound, e:
+                    return tag.a(label, class_='attachment',
+                                 href=raw_href + params,
+                                 title=get_resource_name(self.env, attachment))
+                href = get_resource_url(self.env, attachment, formatter.href)
+                title = get_resource_name(self.env, attachment)
+                img = tag.img(src=formatter.href.chrome('common/download.png'),
+                              alt=_("Download"))
+                return tag(tag.a(label, class_='attachment', title=title,
+                                 href=href + params),
+                           tag.span(" ",
+                                    tag.a(img, class_='trac-rawlink',
+                                          href=raw_href + params,
+                                          title=_("Download")),
+                                    class_="noprint"))
+            except ResourceNotFound:
                 pass
             # FIXME: should be either:
             #
@@ -769,7 +857,7 @@ class LegacyAttachmentPolicy(Component):
             return
         legacy_action = perm_map.get(resource.parent.realm)
         if legacy_action:
-            decision = legacy_action in perm
+            decision = legacy_action in perm(resource.parent)
             if not decision:
                 self.env.log.debug('LegacyAttachmentPolicy denied %s '
                                    'access to %s. User needs %s' %
@@ -781,3 +869,121 @@ class LegacyAttachmentPolicy(Component):
                         resource, perm)
                 if decision is not None:
                     return decision
+
+
+class AttachmentAdmin(Component):
+    """trac-admin command provider for attachment administration."""
+    
+    implements(IAdminCommandProvider)
+    
+    # IAdminCommandProvider methods
+    
+    def get_admin_commands(self):
+        yield ('attachment list', '<realm:id>',
+               """List attachments of a resource
+               
+               The resource is identified by its realm and identifier.""",
+               self._complete_list, self._do_list)
+        yield ('attachment add', '<realm:id> <path> [author] [description]',
+               """Attach a file to a resource
+               
+               The resource is identified by its realm and identifier. The
+               attachment will be named according to the base name of the file.
+               """,
+               self._complete_add, self._do_add)
+        yield ('attachment remove', '<realm:id> <name>',
+               """Remove an attachment from a resource
+               
+               The resource is identified by its realm and identifier.""",
+               self._complete_remove, self._do_remove)
+        yield ('attachment export', '<realm:id> <name> [destination]',
+               """Export an attachment from a resource to a file or stdout
+               
+               The resource is identified by its realm and identifier. If no
+               destination is specified, the attachment is output to stdout.
+               """,
+               self._complete_export, self._do_export)
+    
+    def get_realm_list(self):
+        rs = ResourceSystem(self.env)
+        return PrefixList([each + ":" for each in rs.get_known_realms()])
+    
+    def split_resource(self, resource):
+        result = resource.split(':', 1)
+        if len(result) != 2:
+            raise AdminCommandError(_("Invalid resource identifier '%(id)s'",
+                                      id=resource))
+        return result
+    
+    def get_attachment_list(self, resource):
+        (realm, id) = self.split_resource(resource)
+        return [a.filename for a in Attachment.select(self.env, realm, id)]
+    
+    def _complete_list(self, args):
+        if len(args) == 1:
+            return self.get_realm_list()
+    
+    def _complete_add(self, args):
+        if len(args) == 1:
+            return self.get_realm_list()
+        elif len(args) == 2:
+            return get_dir_list(args[1])
+    
+    def _complete_remove(self, args):
+        if len(args) == 1:
+            return self.get_realm_list()
+        elif len(args) == 2:
+            return self.get_attachment_list(args[0])
+    
+    def _complete_export(self, args):
+        if len(args) < 3:
+            return self._complete_remove(args)
+        elif len(args) == 3:
+            return get_dir_list(args[2])
+    
+    def _do_list(self, resource):
+        (realm, id) = self.split_resource(resource)
+        print_table([(a.filename, pretty_size(a.size), a.author,
+                      format_datetime(a.date, console_datetime_format),
+                      a.description)
+                     for a in Attachment.select(self.env, realm, id)],
+                    [_('Name'), _('Size'), _('Author'), _('Date'),
+                     _('Description')])
+    
+    def _do_add(self, resource, path, author='trac', description=''):
+        (realm, id) = self.split_resource(resource)
+        attachment = Attachment(self.env, realm, id)
+        attachment.author = author
+        attachment.description = description
+        f = open(path, 'rb')
+        try:
+            attachment.insert(os.path.basename(path), f, os.path.getsize(path))
+        finally:
+            f.close()
+    
+    def _do_remove(self, resource, name):
+        (realm, id) = self.split_resource(resource)
+        attachment = Attachment(self.env, realm, id, name)
+        attachment.delete()
+    
+    def _do_export(self, resource, name, destination=None):
+        (realm, id) = self.split_resource(resource)
+        attachment = Attachment(self.env, realm, id, name)
+        if destination is not None:
+            if os.path.isdir(destination):
+                destination = os.path.join(destination, name)
+            if os.path.isfile(destination):
+                raise AdminCommandError(_("File '%(name)s' exists",
+                                          name=destination))
+        input = attachment.open()
+        try:
+            output = (destination is None) and sys.stdout \
+                                           or open(destination, "wb")
+            try:
+                shutil.copyfileobj(input, output)
+            finally:
+                if destination is not None:
+                    output.close()
+        finally:
+            input.close()
+
