@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2005-2008 Edgewall Software
+# Copyright (C) 2005-2009 Edgewall Software
 # Copyright (C) 2005-2006 Emmanuel Blot <emmanuel.blot@free.fr>
 # All rights reserved.
 #
@@ -16,7 +16,6 @@
 # (lsmithson@open-networks.co.uk) extensible Python SMTP Server
 #
 
-from trac.core import TracError
 from trac.util.datefmt import utc
 from trac.ticket.model import Ticket
 from trac.ticket.notification import TicketNotifyEmail
@@ -29,7 +28,6 @@ from datetime import datetime
 import os
 import quopri
 import re
-import time
 import unittest
 
 SMTP_TEST_PORT = 7000 + os.getpid() % 1000
@@ -42,7 +40,7 @@ class NotificationTestCase(unittest.TestCase):
     
     def setUp(self):
         self.env = EnvironmentStub(default_data=True)
-        self.env.config.set('project','name', 'TracTest')
+        self.env.config.set('project', 'name', 'TracTest')
         self.env.config.set('notification', 'smtp_enabled', 'true')
         self.env.config.set('notification', 'always_notify_owner', 'true')
         self.env.config.set('notification', 'always_notify_reporter', 'true')
@@ -57,11 +55,12 @@ class NotificationTestCase(unittest.TestCase):
     def tearDown(self):
         """Signal the notification test suite that a test is over"""
         notifysuite.tear_down()
+        self.env.reset_db()
 
     def test_recipients(self):
         """To/Cc recipients"""
         ticket = Ticket(self.env)
-        ticket['reporter'] = '"Joe User" <joe.user@example.org>'
+        ticket['reporter'] = '"Joe User" < joe.user@example.org >'
         ticket['owner']    = 'joe.user@example.net'
         ticket['cc']       = 'joe.user@example.com, joe.bar@example.org, ' \
                              'joe.bar@example.net'
@@ -142,16 +141,17 @@ class NotificationTestCase(unittest.TestCase):
         """Date format compliance (RFC822) 
            we do not support 'military' format""" 
         date_str = r"^((?P<day>\w{3}),\s*)*(?P<dm>\d{2})\s+" \
-                   r"(?P<month>\w{3})\s+(?P<year>200\d)\s+" \
+                   r"(?P<month>\w{3})\s+(?P<year>\d{4})\s+" \
                    r"(?P<hour>\d{2}):(?P<min>[0-5][0-9])" \
                    r"(:(?P<sec>[0-5][0-9]))*\s" \
                    r"((?P<tz>\w{2,3})|(?P<offset>[+\-]\d{4}))$"
         date_re = re.compile(date_str)
         # python time module does not detect incorrect time values
-        days = ['Mon','Tue','Wed','Thu','Fri','Sat','Sun']
-        months = ['Jan','Feb','Mar','Apr','May','Jun', \
-                  'Jul','Aug','Sep','Oct','Nov','Dec']
-        tz = ['UT','GMT','EST','EDT','CST','CDT','MST','MDT''PST','PDT']
+        days = ['Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun']
+        months = ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', \
+                  'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec']
+        tz = ['UT', 'GMT', 'EST', 'EDT', 'CST', 'CDT', 'MST', 'MDT',
+              'PST', 'PDT']
         ticket = Ticket(self.env)
         ticket['reporter'] = '"Joe User" <joe.user@example.org>'
         ticket['summary'] = 'This is a summary'
@@ -165,9 +165,9 @@ class NotificationTestCase(unittest.TestCase):
         self.failIf(not mo)
         if mo.group('day'):
             self.failIf(mo.group('day') not in days)
-        self.failIf(int(mo.group('dm')) not in range(1,32))
+        self.failIf(int(mo.group('dm')) not in range(1, 32))
         self.failIf(mo.group('month') not in months)
-        self.failIf(int(mo.group('hour')) not in range(0,24))
+        self.failIf(int(mo.group('hour')) not in range(0, 24))
         if mo.group('tz'):
             self.failIf(mo.group('tz') not in tz)
 
@@ -374,7 +374,7 @@ class NotificationTestCase(unittest.TestCase):
 
     def test_multiline_header(self):
         """Encoded headers split into multiple lines"""
-        self.env.config.set('notification','mime_encoding', 'qp')
+        self.env.config.set('notification', 'mime_encoding', 'qp')
         ticket = Ticket(self.env)
         ticket['reporter'] = 'joe.user@example.org'
         # Forces non-ascii characters
@@ -391,7 +391,7 @@ class NotificationTestCase(unittest.TestCase):
 
     def test_mimebody_b64(self):
         """MIME Base64/utf-8 encoding"""
-        self.env.config.set('notification','mime_encoding', 'base64')
+        self.env.config.set('notification', 'mime_encoding', 'base64')
         ticket = Ticket(self.env)
         ticket['reporter'] = 'joe.user@example.org'
         ticket['summary'] = u'This is a long enough summary to cause Trac ' \
@@ -402,7 +402,7 @@ class NotificationTestCase(unittest.TestCase):
 
     def test_mimebody_qp(self):
         """MIME QP/utf-8 encoding"""
-        self.env.config.set('notification','mime_encoding', 'qp')
+        self.env.config.set('notification', 'mime_encoding', 'qp')
         ticket = Ticket(self.env)
         ticket['reporter'] = 'joe.user@example.org'
         ticket['summary'] = u'This is a long enough summary to cause Trac ' \
@@ -411,14 +411,24 @@ class NotificationTestCase(unittest.TestCase):
         self._validate_mimebody((quopri, 'quoted-printable', 'utf-8'),
                                 ticket, True)
 
-    def test_mimebody_none(self):
-        """MIME None/ascii encoding"""
-        self.env.config.set('notification','mime_encoding', 'none')
+    def test_mimebody_none_7bit(self):
+        """MIME None encoding resulting in 7bit"""
+        self.env.config.set('notification', 'mime_encoding', 'none')
         ticket = Ticket(self.env)
         ticket['reporter'] = 'joe.user'
         ticket['summary'] = u'This is a summary'
         ticket.insert()
-        self._validate_mimebody((None, '7bit', 'ascii'), \
+        self._validate_mimebody((None, '7bit', 'utf-8'), \
+                                ticket, True)
+
+    def test_mimebody_none_8bit(self):
+        """MIME None encoding resulting in 8bit"""
+        self.env.config.set('notification', 'mime_encoding', 'none')
+        ticket = Ticket(self.env)
+        ticket['reporter'] = 'joe.user'
+        ticket['summary'] = u'This is a summary for Jöe Usèr'
+        ticket.insert()
+        self._validate_mimebody((None, '8bit', 'utf-8'), \
                                 ticket, True)
 
     def test_md5_digest(self):
@@ -439,7 +449,7 @@ class NotificationTestCase(unittest.TestCase):
         """No-self-notification option"""
         def _test_updater(disable):
             if disable:
-                self.env.config.set('notification','always_notify_updater',
+                self.env.config.set('notification', 'always_notify_updater',
                                     'false')
             ticket = Ticket(self.env)
             ticket['reporter'] = 'joe.user@example.org'
@@ -600,12 +610,12 @@ class NotificationTestCase(unittest.TestCase):
                 self.failIf(footer > 3)
                 # check ticket link
                 if line[:11] == 'Ticket URL:':
-                    self.assertEqual(line[12:].strip(),
-                                     "<%s>" % ticket['link'].strip())
+                    ticket_link = self.env.abs_href.ticket(ticket.id)
+                    self.assertEqual(line[12:].strip(), "<%s>" % ticket_link)
                 # note project title / URL are not validated yet
 
         # ticket properties which are not expected in the banner
-        xlist = ['summary', 'description', 'link', 'comment', 'new']
+        xlist = ['summary', 'description', 'comment', 'time', 'changetime']
         # check banner content (field exists, msg value matches ticket value)
         for p in [prop for prop in ticket.values.keys() if prop not in xlist]:
             self.failIf(not props.has_key(p))
@@ -615,6 +625,31 @@ class NotificationTestCase(unittest.TestCase):
             else:
                 self.failIf(props[p] != ticket[p])
 
+    def test_notification_does_not_alter_ticket_instance(self):
+        ticket = Ticket(self.env)
+        ticket['summary'] = 'My Summary'
+        ticket['description'] = 'Some description'
+        ticket.insert()
+        tn = TicketNotifyEmail(self.env)
+        tn.notify(ticket, newticket=True)
+        self.assertNotEqual(None, notifysuite.smtpd.get_message())
+        self.assertEqual('My Summary', ticket['summary'])
+        self.assertEqual('Some description', ticket['description'])
+        valid_fieldnames = set([f['name'] for f in ticket.fields])
+        current_fieldnames = set(ticket.values.keys())
+        self.assertEqual(set(), current_fieldnames - valid_fieldnames)
+
+    def test_notification_get_message_id_unicode(self):
+        ticket = Ticket(self.env)
+        ticket['summary'] = 'My Summary'
+        ticket['description'] = 'Some description'
+        ticket.insert()
+        self.env.config.set('project', 'url', u"пиво Müller ")
+        tn = TicketNotifyEmail(self.env)
+        tn.ticket = ticket
+        tn.get_message_id('foo')
+        
+
 
 class NotificationTestSuite(unittest.TestSuite):
     """Thin test suite wrapper to start and stop the SMTP test server"""
