@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2003-2008 Edgewall Software
+# Copyright (C) 2003-2009 Edgewall Software
 # Copyright (C) 2003-2005 Jonas Borgström <jonas@edgewall.com>
 # Copyright (C) 2004-2005 Christopher Lenz <cmlenz@gmx.de>
 # Copyright (C) 2005-2006 Christian Boos <cboos@neuf.fr>
@@ -20,8 +20,6 @@
 from datetime import datetime, timedelta
 import pkg_resources
 import re
-import time
-from urlparse import urlparse
 
 from genshi.builder import tag
 
@@ -30,11 +28,11 @@ from trac.core import *
 from trac.mimeview import Context
 from trac.perm import IPermissionRequestor
 from trac.timeline.api import ITimelineEventProvider
-from trac.util.compat import sorted
+from trac.util import as_int
 from trac.util.datefmt import format_date, format_datetime, parse_date, \
-                              to_timestamp, utc, pretty_timedelta
-from trac.util.text import to_unicode
-from trac.util.translation import _
+                              to_utimestamp, utc, pretty_timedelta
+from trac.util.text import exception_to_unicode, to_unicode
+from trac.util.translation import _, tag_
 from trac.web import IRequestHandler, IRequestFilter
 from trac.web.chrome import add_link, add_stylesheet, prevnext_nav, Chrome, \
                             INavigationContributor, ITemplateProvider
@@ -65,6 +63,8 @@ class TimelineModule(Component):
         specific event providers, see their own documentation.
         (''Since 0.11'')""")
 
+    _authors_pattern = re.compile(r'(-)?(?:"([^"]*)"|\'([^\']*)\'|([^\s]+))')
+
     # INavigationContributor methods
 
     def get_active_navigation_item(self, req):
@@ -89,15 +89,18 @@ class TimelineModule(Component):
         req.perm.assert_permission('TIMELINE_VIEW')
 
         format = req.args.get('format')
-        maxrows = int(req.args.get('max', 0))
+        maxrows = int(req.args.get('max', format == 'rss' and 50 or 0))
 
         # Parse the from date and adjust the timestamp to the last second of
         # the day
         fromdate = today = datetime.now(req.tz)
         precisedate = precision = None
         if 'from' in req.args:
-            precisedate = parse_date(req.args.get('from'), req.tz)
-            fromdate = precisedate
+            # Acquire from date only from non-blank input
+            reqfromdate = req.args['from'].strip()
+            if reqfromdate:
+                precisedate = parse_date(reqfromdate, req.tz)
+                fromdate = precisedate
             precision = req.args.get('precision', '')
             if precision.startswith('second'):
                 precision = timedelta(seconds=1)
@@ -107,38 +110,44 @@ class TimelineModule(Component):
                 precision = timedelta(hours=1)
             else:
                 precision = None
-        fromdate = fromdate.replace(hour=23, minute=59, second=59)
-        try:
-            daysback = int(req.args.get('daysback', ''))
-        except ValueError:
-            try:
-                daysback = int(req.session.get('timeline.daysback', ''))
-            except ValueError:
-                daysback = self.default_daysback
+        fromdate = fromdate.replace(hour=23, minute=59, second=59,
+                                    microsecond=999999)
+
+        daysback = as_int(req.args.get('daysback'),
+                          format == 'rss' and 90 or None)
+        if daysback is None:
+            daysback = as_int(req.session.get('timeline.daysback'), None)
+        if daysback is None:
+            daysback = self.default_daysback
         daysback = max(0, daysback)
         if self.max_daysback >= 0:
             daysback = min(self.max_daysback, daysback)
 
+        authors = req.args.get('authors')
+        if authors is None and format != 'rss':
+            authors = req.session.get('timeline.authors')
+        authors = (authors or '').strip()
+
         data = {'fromdate': fromdate, 'daysback': daysback,
-                'today': format_date(today),
-                'yesterday': format_date(today - timedelta(days=1)),
+                'authors': authors,
+                'today': format_date(today, tzinfo=req.tz),
+                'yesterday': format_date(today - timedelta(days=1),
+                                         tzinfo=req.tz),
                 'precisedate': precisedate, 'precision': precision,
                 'events': [], 'filters': [],
                 'abbreviated_messages': self.abbreviated_messages}
 
         available_filters = []
         for event_provider in self.event_providers:
-            available_filters += event_provider.get_timeline_filters(req)
+            available_filters += event_provider.get_timeline_filters(req) or []
 
-        filters = []
         # check the request or session for enabled filters, or use default
-        for test in (lambda f: f[0] in req.args,
-                     lambda f: req.session.get('timeline.filter.%s' % f[0],
-                                               '') == '1',
-                     lambda f: len(f) == 2 or f[2]):
-            if filters:
-                break
-            filters = [f[0] for f in available_filters if test(f)]
+        filters = [f[0] for f in available_filters if f[0] in req.args]
+        if not filters and format != 'rss':
+            filters = [f[0] for f in available_filters
+                       if req.session.get('timeline.filter.' + f[0]) == '1']
+        if not filters:
+            filters = [f[0] for f in available_filters if len(f) == 2 or f[2]]
 
         # save the results of submitting the timeline form to the session
         if 'update' in req.args:
@@ -152,13 +161,27 @@ class TimelineModule(Component):
         stop = fromdate
         start = stop - timedelta(days=daysback + 1)
 
+        # create author include and exclude sets
+        include = set()
+        exclude = set()
+        for match in self._authors_pattern.finditer(authors):
+            name = (match.group(2) or match.group(3) or match.group(4)).lower()
+            if match.group(1):
+                exclude.add(name)
+            else:
+                include.add(name)
+        
         # gather all events for the given period of time
         events = []
         for provider in self.event_providers:
             try:
                 for event in provider.get_timeline_events(req, start, stop,
-                                                          filters):
-                    events.append(self._event_data(provider, event))
+                                                          filters) or []:
+                    # Check for 0.10 events
+                    author = (event[len(event) < 6 and 2 or 4] or '').lower()
+                    if (not include or author in include) \
+                       and not author in exclude:
+                        events.append(self._event_data(provider, event))
             except Exception, e: # cope with a failure of that provider
                 self._provider_failure(e, req, provider, filters,
                                        [f[0] for f in available_filters])
@@ -170,29 +193,24 @@ class TimelineModule(Component):
 
         data['events'] = events
         
-
         if format == 'rss':
-            # Get the email addresses of all known users
-            email_map = {}
-            if Chrome(self.env).show_email_addresses:
-                for username, name, email in self.env.get_known_users():
-                    if email:
-                        email_map[username] = email
-            data['email_map'] = email_map
+            data['email_map'] = Chrome(self.env).get_email_map()
             rss_context = Context.from_request(req, absurls=True)
             rss_context.set_hints(wiki_flavor='html', shorten_lines=False)
             data['context'] = rss_context
             return 'timeline.rss', data, 'application/rss+xml'
         else:
             req.session['timeline.daysback'] = daysback
-            html_context = Context.from_request(req, absurls=True)
+            req.session['timeline.authors'] = authors
+            html_context = Context.from_request(req)
             html_context.set_hints(wiki_flavor='oneliner', 
                                    shorten_lines=self.abbreviated_messages)
             data['context'] = html_context
 
         add_stylesheet(req, 'common/css/timeline.css')
         rss_href = req.href.timeline([(f, 'on') for f in filters],
-                                     daysback=90, max=50, format='rss')
+                                     daysback=90, max=50, authors=authors,
+                                     format='rss')
         add_link(req, 'alternate', rss_href, _('RSS Feed'),
                  'application/rss+xml', 'rss')
 
@@ -204,15 +222,17 @@ class TimelineModule(Component):
         previous_start = format_date(fromdate - timedelta(days=daysback+1),
                                      format='%Y-%m-%d', tzinfo=req.tz)
         add_link(req, 'prev', req.href.timeline(from_=previous_start,
+                                                authors=authors,
                                                 daysback=daysback),
-                 _('Previous period'))
+                 _('Previous Period'))
         if today - fromdate > timedelta(days=0):
             next_start = format_date(fromdate + timedelta(days=daysback+1),
                                      format='%Y-%m-%d', tzinfo=req.tz)
             add_link(req, 'next', req.href.timeline(from_=next_start,
+                                                    authors=authors,
                                                     daysback=daysback),
-                     _('Next period'))
-        prevnext_nav(req, 'Period')
+                     _('Next Period'))
+        prevnext_nav(req, _('Previous Period'), _('Next Period'))
         
         return 'timeline.html', data, None
 
@@ -292,11 +312,9 @@ class TimelineModule(Component):
                 kind, date, author, data = event
             render = lambda field, context: \
                     provider.render_timeline_event(context, field, event)
-        if isinstance(date, datetime):
-            dateuid = to_timestamp(date)
-        else:
-            dateuid = date
+        if not isinstance(date, datetime):
             date = datetime.fromtimestamp(date, utc)
+        dateuid = to_utimestamp(date)
         return {'kind': kind, 'author': author, 'date': date,
                 'dateuid': dateuid, 'render': render, 'event': event,
                 'data': data, 'provider': provider}
@@ -307,20 +325,29 @@ class TimelineModule(Component):
         At the same time, the message will contain a link to the timeline
         without the filters corresponding to the guilty event provider `ep`.
         """
-        ep_name, exc_name = [i.__class__.__name__ for i in (ep, exc)]
-        self.log.exception('Timeline event provider %s failed', ep_name)
+        self.log.error('Timeline event provider failed: %s', 
+                       exception_to_unicode(exc, traceback=True))
 
-        guilty_filters = [f[0] for f in ep.get_timeline_filters(req)]
-        guilty_kinds = [f[1] for f in ep.get_timeline_filters(req)]
-        other_filters = [f for f in current_filters if not f in guilty_filters]
+        ep_kinds = dict((f[0], f[1])
+                        for f in ep.get_timeline_filters(req) or [])
+        ep_filters = set(ep_kinds.keys())
+        current_filters = set(current_filters)
+        other_filters = set(current_filters) - ep_filters
         if not other_filters:
-            other_filters = [f for f in all_filters if not f in guilty_filters]
+            other_filters = set(all_filters) -  ep_filters
         args = [(a, req.args.get(a)) for a in ('from', 'format', 'max',
                                                'daysback')]
-        href = req.href.timeline(args+[(f, 'on') for f in other_filters])
+        href = req.href.timeline(args + [(f, 'on') for f in other_filters])
+        # TRANSLATOR: ...want to see the 'other kinds of events' from... (link)
+        other_events = tag.a(_('other kinds of events'), href=href)
         raise TracError(tag(
-            tag.p(', '.join(guilty_kinds),
-                  ' event provider (', tag.tt(ep_name), ') failed:', tag.br(),
-                  exc_name, ': ', to_unicode(exc), class_='message'),
-            tag.p('You may want to see the other kind of events from the ',
-                  tag.a('Timeline', href=href))))
+            tag.p(tag_("Event provider %(name)s failed for filters "
+                       "%(kinds)s: ",
+                       name=tag.tt(ep.__class__.__name__),
+                       kinds=', '.join('"%s"' % ep_kinds[f] for f in 
+                                       current_filters & ep_filters)),
+                  tag.b(exception_to_unicode(exc)), class_='message'),
+            tag.p(tag_("You may want to see the %(other_events)s from the "
+                       "Timeline or notify your Trac administrator about the "
+                       "error (detailed information was written to the log).",
+                       other_events=other_events))))
